function [COEFF,OUT,DATAOUT,parametout] = RHA_optim_final(NMAX,TAB)
%%% Rectangular Harmonic Analysis
%%% RHA_dF_optim.m: [COEFF,stddf,OUT] = RHA_dF_optim(NMAX,TAB)
%%% Inputs: NMAX = maximum rectangular harmonic degree
%%%         TAB  = can be data to be modelled in the correct format: see
%%%         loading data below
%%% Output: COEFF = coefficients of the model in RHA
%%%         OUT = [Lat Long Alt dF dBr dBtheta dBphi Bcore_r
%%%         Bcore_theta Bcore_phi Fcore]; % on the
%%%         regular grid.
%%%         DATAOUT = [Lat Long Alt dF dBr dBtheta dBphi Bcore_r
%%%         Bcore_theta Bcore_phi Fcore dF_original dF_residuals weights]; % on the real
%%%         data location.
%%%         parametout: structure containing output grid information

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% E. THEBAULT, Centre National de la Recherche Scientifique (CNRS)
%%% Universite Clermont Auvergne, IRD, OPGC, Laboratoire Magmas et Volcans, Clermont-Ferrand, France
%%% Version 1.0, March 2023 
%%% Version 1.1, September 2023: modification of plots and appending the
%%% indices to the coefficients.
%%% Version 2.0, October 2023: input is now a structure to avoid confusion and simplify the core field computation 
%%% Please cite: 
%%% Thebault, E., & Gailler, L. S. (2023). A Quick-Look Software for In Situ 
%%% Magnetic Field Modeling from Onboard Unmanned Aircraft Vehicles (UAVs) 
%%% Measurements. Remote Sensing, 15(18), 4549. 
%%% https://doi.org/10.3390/rs15184549

%%% 1) GENERAL COMMENTS AND WARNING
%%% Input format of data file:
%%% TAB is a structure containing at least TAB.Lat = Latitude; TAB.Long =
%%% Longitude; TAB.Alt = Altitude in [km]; TAB.F = Scalar intensity (or
%%% anomaly). The Lat and Long are in decimal degrees.
%%% Two optional fields can be added: TAB.time = decimal year of the data;
%%% TAB.priorw = a priori data quality independently fro each other.
%%% if not available additionnal field should be inexistant (not empty like TAB.time = [];) 

%%% NOTE 1: The core field model is computed according to the CHAOS model: 
%%% http://www.spacecenter.dk/files/magnetic-models/CHAOS-7/

%%% NOTE 2: The algorithm relies on an inverse problem and memory issues 
%%% may appear for large datasets or high RHA degree expansion, depending
%%% on the computer capabilities.

%%% NOTE 3: Matlab does not like large dataset (~1e6) to be plotted and some
%%% issues may be encountered if plotts is set to 1.

%1) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% variables
Rt = 6371.2;
Lanczos = 1; %% 1 for sigma forward computation, 0 for regular forward computation
plotts = 1; %% if plotts = 0, no plot. 
loopmax = 10;
CHAOS = 'CHAOS-7.13.mat';

if nargin < 2
disp(['Uncomment and choose a toy file.');
%%% synthetic data files described in the article
%load('SYN-UAVS-NOISED.mat');
%load('SYN-UAVS-PERFECT.mat');
%load('SYN-AERO-PERFECT.mat');

%%% Additional dataset containing the time and prioriw fields
% load('SYN-AERO-PERFECT_F_TIME_WEIGHTS.mat');
end

a_weights = isfield(TAB,'priorw');
a_time = isfield(TAB,'time');

%%% get the number of data
nd = size(TAB.Lat,1);

if (a_weights == 1) && (a_time == 1)
disp('Data epoch and weighting.');
priorw = TAB.priorw;
decyear_ = TAB.time;
end


if (a_weights == 0) && (a_time == 0)
disp('No data epoch, no weighting.');
priorw = ones(nd,1);
vecyear = datevec(now);
decyear_ = (vecyear(1)+vecyear(2)/12)*ones(nd,1);
end

if (a_weights == 0) && (a_time == 1)
disp('No data weighting, epoch included.');
decyear_ = TAB.time;
priorw = ones(nd,1);
end

if (a_time == 0) && (a_weights == 1)
disp('No data epoch, weights included.');
priorw = TAB.priorw;
priorw = priorw/max(priorw); % normalize
vecyear = datevec(now);
decyear_ = (vecyear(1)+vecyear(2)/12)*ones(nd,1);
end

% Retrieve latitude and longitude
Long =     TAB.Long;
Lat  = TAB.Lat;
CLat =  90-Lat;
Fo = TAB.F;
Alt = TAB.Alt; 
DATAOUT = [Lat Long Alt];

%2) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% convert the data into a local spherical frame
% geometric centre for the  projection
Long0 = (max(Long)+min(Long))/2;
CLat0 = 90-(max(Lat)+min(Lat))/2;

% Simple trick to avoid repeated
% data warning adding noise on data location.
% generate level of Noise at 1/1000 of minimum spacing
dLong = abs(diff(Long));
Long = Long+min(dLong(dLong~=0))/100*randn(length(Long),1);
dCLat = abs(diff(CLat));
CLat = CLat+min(dCLat(dCLat~=0))/100*randn(length(CLat),1);

% Rotation on the Sphere and projection on the plane for the x,y,z
% coordinates
[LL, CC]=geoc2cal(Long0,CLat0,Long,CLat);

%3) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Compute main field values
MJD2000 = datenum(2000,1,1);

%%% Load the CHAOS model
 if fopen(CHAOS)~=-1
        load(CHAOS);
 else
  error('CHAOS model not found ( go to http://www.spacecenter.dk/files/magnetic-models/CHAOS-7/)')   
 end
 
 %%% Validity of CHAOS
 epoch_max = max(pp.breaks);
 epoch_min = min(pp.breaks);
 
%%% convert time into Modified Julian Day 2000: neglect leap year
t = datenum(floor(decyear_),1,1)+365*(decyear_-floor(decyear_))-MJD2000;
Lsup = t > epoch_max;
t(Lsup) = epoch_max;
Linf = t < epoch_min;

if sum(Linf)~=0
    error('Minimum epoch outside the range of CHAOS found.');
end


%%% check out lateral extension and time period to avoid unnecessary computation
dist_max = max(CC); % max distance between the center and the edge
time_reduc = median(t);
if numel(unique(t)) == 1 && dist_max < 1
    %%% compute core field at the center and assumes equal over the
    %%% survey
    
    disp(' SMALL SURVEY EXTENSION, ASSUME CONSTANT CORE FIELD: FASTER');
    B_MF  = ESTIMATE_CORE_FIELD(time_reduc, (Alt(1)+Rt), CLat0*pi/180, Long0*pi/180, pp);
    B_MF = repmat(B_MF,nd,1);
    
else
    B_MF  = ESTIMATE_CORE_FIELD(t, (Alt+Rt), CLat*pi/180, Long*pi/180, pp);
end


BX = -B_MF(:,2);
BY =  B_MF(:,3);
BZ = -B_MF(:,1);
Fcoreo = B_MF(:,4);

%%% clean up variables generated by the main field computation
clear t pp_CAL params options mt model_ext model_Euler g CHAOS

%4) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Rotations and projections.
%%% Ideally dFo should be Fo-Fcoreo, however the regional trend might cause
%%% numerical issues in the RHA modelling, also input data might be reduced already,
%%% check out case by case.
if mean(abs(Fo)) < 1000;
    disp('Already in the form of anomaly intensity.');
    dFo = Fo;
   else
    disp('Correcting for the main intensity field.');
    dFo = Fo-Fcoreo;
end

%%% Local spherical to Lambert rectangular projection
A = prj_lamA([LL CC],'forward');


%%% Spherical to plane transformation of vector field components
CoreField = Sph2rectangular(Long0,CLat0,Long,CLat,[BX BY BZ],'forward');
CoreField(:,4) = Fcoreo;

%%% x is not abscissa but defined northward, y east-west, z is
%%% positive downward.
xo = -A(:,1);
yo =  A(:,2);
zo = -(Alt-A(:,3));

%%% this altitude with resepct to the Earth's surface is used for the final map
alt_min = median(Alt);
%alt_min = min(Alt);

%%% add a small empty border to push back the ringing as a fuction of NMAX
xmax = max(abs(xo))+max(abs(xo))/(2*NMAX);
ymax = max(abs(yo))+max(abs(yo))/(2*NMAX);

DATAOUT = [DATAOUT zeros(nd,4) B_MF zeros(nd,2)];

clear BX BY BZ B_MF LL CC dLong dCLat Fo TAB

%5) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% RHA Inversion
DXp = 2*xmax;
DYp = 2*ymax;
DZp = abs(max(zo)-min(zo));
DHp = sqrt((DXp^2+DYp^2));

%%% Build the design matrix for RHA
[MAT,IND,MATXo,MATYo,MATZo] = design_RHA(xo,yo,zo,CoreField,DXp,DYp,NMAX);

%%% keep non null basis function
LP = sum(MAT,1) ~= 0;

%%%
disp(['******************************************************************']);
disp(['Rectangular Harmonic Analysis developed to NMAX =  ' num2str(NMAX)]);
disp(['*****involves =  ' num2str(size(MAT,2)) ' parameters.']);
disp(['*****involves =  ' num2str(size(MAT,1)) ' data points.']);
disp(['Maximum altitude difference between data: ' num2str(DZp) ' km.']);
disp(['Minimum resolved wavelength along North : ' num2str((DXp)/NMAX) ' km.']);
disp(['Minimum resolved wavelength along Est : ' num2str((DYp)/NMAX) ' km.']);
disp([' ---> Minimum horizontal wavelength    : ' num2str((DHp)/NMAX) ' km.']);
disp('----------------------------------------------------------------');
disp('--------- INVERSION AND FIELD COMPONENT RECOVERY----------------');
disp('----------------------------------------------------------------');

dFo = dFo-mean(dFo);

%%% Iterative weighting scheme with Huber Weigths: note, alternative
%%% robustfit from matlab could be used but requires the optimization
%%% toolbox (and is slower) and does not allow GSVD.

misfit = 1e9;
eps = misfit;
epslim = 1e-2;
iik = 1;
eigenratio = 1e-4;
huber = 1.345;

%%% generate sparse matrix for the weights (diagonal matrix).
indi = 1:size(MAT,1);
WE = sparse(indi,indi,priorw./max(priorw));
COEFP(:,1) = zeros(sum(LP),1);

for iik=1:loopmax
    ATWA = MAT(:,LP)'*WE*MAT(:,LP);
    ATW  = MAT(:,LP)'*WE*dFo;
    
    %%% Generalized SVD inversion
    COEFP(:,iik+1) = PreCondLight_2_threshold(ATWA,ATW,eigenratio);
    disf = (MAT(:,LP)*COEFP(:,iik+1)-dFo);
    
    stde(iik,1) = std(disf); % standard estimate
    stde(iik,2) = mad(disf)/0.6745; % robust estimate
    
    %%% rebuild the weighting matrix for deviant data according to huber
    rr = disf./(huber*stde(iik,2)*sqrt(priorw));
    priorw = 1./max(1,abs(rr));
    WE = sparse(indi,indi,priorw);
     
    %%% breaking rule when the convergence is reached within epslim
    if (norm(COEFP(:,iik+1)-COEFP(:,iik),2))/norm(COEFP(:,iik+1)) < epslim;
        disp(['Coefficients converged after ' num2str(iik) ' iterations.']);
       break;
    end
    iik = iik+1;
end

COEFF(LP,1) = COEFP(:,iik); clear COEFP;
dFesto = MAT*COEFF;
Xesto = MATXo*COEFF;
Yesto = MATYo*COEFF;
Zesto = MATZo*COEFF;

%%% Back projection from plane to spherical of the predictions on original
%%% locations.
Bcrust = Sph2rectangular(Long0,CLat0,Long,CLat,[Xesto Yesto Zesto],'reverse');
DATAOUT(:,4) = dFesto;
DATAOUT(:,5) = -Bcrust(:,3); 
DATAOUT(:,6) = -Bcrust(:,1);
DATAOUT(:,7) =  Bcrust(:,2);
DATAOUT(:,12) = dFo;
DATAOUT(:,13) = dFo-dFesto;
DATAOUT(:,14) = priorw;

dres = std(dFesto-dFo);
disp('*******************************************************************')
disp(['Final iterative misfit dF = ' num2str(dres) 'nT ; unbiased estimate : ' num2str(stde(iik-1,2)) ' nT.'])

%6) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Forward computation on a regular grid for the maps
xmin = -max([abs(min(xo)) max(xo)]);
xmax =  max([abs(min(xo)) max(xo)]);
ymin = -max([abs(min(yo)) max(yo)]);
ymax =  max([abs(min(yo)) max(yo)]);
step = min([xmax ymax])/(4*NMAX);
xi = xmin:step:xmax;
yi = ymin:step:ymax;
[XX,YY] = meshgrid(xi,yi);

disp(['Output regularly spaced grid : ' num2str(length(xi)) ' East ;' num2str(length(yi)) ' North.']);
%%% unit area of the mesh in meters
lengthstep = (step*1e3);

%%% reshape grids
x = reshape(XX,numel(XX),1);
y = reshape(YY,numel(YY),1);

%%% Identify inside points only (no extrapolation outside the dominion of
%%% data.
FF = scatteredInterpolant(yo,xo,dFo,'natural','none');
dFtrue = FF(y,x);
LN = isfinite(dFtrue);

%%% project back the cartesian regular grid into the local geocentric reference frame
%%% centered on mean Latitude and longitude
MM = prj_lamA([-x y],'reverse');

%%% Convert back locations into the true geocentric reference frame.
[Longi,CLati] = cal2geoc(Long0,CLat0,MM(:,1),MM(:,2));
A = prj_lamA([MM(:,1) MM(:,2)],'forward');
z_r = -(alt_min-A(:,3));

%%% Compute the core field for projection
B_MF  = ESTIMATE_CORE_FIELD(time_reduc*ones(length(A),1), alt_min*ones(length(A),1)+Rt, CLati*pi/180, Longi*pi/180, pp);
Bcore_sp = [-B_MF(:,2) B_MF(:,3) -B_MF(:,1)];

%%% same as before: projection of the core field from spherical to planar.
Bcore_cr = Sph2rectangular(Long0,CLat0,Longi,CLati,Bcore_sp,'forward');
Xcorer =  Bcore_cr(:,1);
Ycorer =  Bcore_cr(:,2);
Zcorer =  Bcore_cr(:,3);
Fcorer =sqrt(Xcorer.^2+Ycorer.^2+Zcorer.^2);
CoreFieldr = [Xcorer Ycorer Zcorer Fcorer];

%%% Build the forward design matrix
[MATN,IND,MATX,MATY,MATZ] = design_RHA(x,y,z_r,CoreFieldr,DXp,DYp,NMAX);


%%% Lanczos weighting
if Lanczos == 1
%%% Lanczos sigma weighting
WW = (sinc(IND(:,1)/(NMAX+1)).*sinc(IND(:,2)/(NMAX+1))).^(1/2);
COEFF(:,1) = COEFF(:,1).*WW;
dFestsigma = MAT*COEFF;
dres = std(dFestsigma-dFo);
disp(['Lanczos sigma misfit dF = ' num2str(dres)])
end

dFr = MATN*COEFF(:,1);
dXr = MATX*COEFF(:,1);
dYr = MATY*COEFF(:,1);
dZr = MATZ*COEFF(:,1);

%%% Back projection from planar to spherical
Bcrust = Sph2rectangular(Long0,CLat0,Longi(LN),CLati(LN),[dXr(LN) dYr(LN) dZr(LN)],'reverse');
%%% Test without back projection
%Bcrust = [dXr(LN) dYr(LN) dZr(LN)];


%%% Append indices to coefficients
COEFF = [COEFF IND];

%%% Sendback useful parameters regarding the regular grid
parametout.x = x(LN);
parametout.y = y(LN);
parametout.Lat = 90-CLati(LN);
parametout.Long = Longi(LN);
parametout.DXp = DXp;
parametout.DYp = DYp;
parametout.NMAX = NMAX;
parametout.dist_dip = lengthstep;

%%% Generate the output in table form for the regular grid
OUT = [Longi(LN) 90-CLati(LN) alt_min*ones(sum(LN),1) dFr(LN) -Bcrust(:,3) -Bcrust(:,1) Bcrust(:,2) B_MF(LN,:)];


%7) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% quick look plot section
if plotts == 1
map = MyColorMap();
%load coast; %%% old matlab versions
load coastlines; lat = coastlat;  long = coastlon; %%% new matlab version

figure;
subplot(2,2,1); scatter(yo(1:1:end),xo(1:1:end),6,dFo(1:1:end),'filled'); colormap(map); caxis([-2*std(dFo) 2*std(dFo)]); colorbar
xlabel('East in [km]'); ylabel('North in [km]'); title('Original scalar anomaly'); axis tight; grid on; 
subplot(2,2,2);scatter(yo(1:1:end),xo(1:1:end),6,dFesto(1:1:end),'filled'); colormap(map); caxis([-2*std(dFo) 2*std(dFo)]); colorbar
xlabel('East in [km]'); ylabel('North in [km]');title('Model scalar anomaly'); axis tight;grid on; 
subplot(2,2,3);scatter(yo(1:1:end),xo(1:1:end),6,dFesto(1:1:end)-dFo(1:1:end),'filled'); colormap(map); caxis([-2*dres 2*dres]); colorbar
xlabel('East in [km]'); ylabel('North in [km]'); title('Misfit dF residuals'); axis tight;grid on; 
subplot(2,2,4);scatter(Longi(LN),90-CLati(LN),6,(dZr(LN)),'filled'); colormap(map); caxis([-2*std(dFo) 2*std(dFo)]); colorbar
hold on; plot(long,lat,'k', 'linewidth',2); xlim([min(Longi(LN)) max(Longi(LN))]); ylim([min(90-CLati(LN)) max(90-CLati(LN))]);
%hold on; plot(Long, 90-CLat,'k.');grid on; 
xlabel('Longitude in  deg.'); ylabel('Latitude in deg.'); title(['Z-component at ' num2str(round(alt_min)) ' km.']); 

figure;scatter(Longi(LN),90-CLati(LN),6,detrend(Bcrust(:,1)),'filled'); colormap(map); caxis([-2*std(dFesto) 2*std(dFesto)]); colorbar
hold on; plot(long,lat,'k', 'linewidth',2); xlim([min(Longi(LN)) max(Longi(LN))]); ylim([min(90-CLati(LN)) max(90-CLati(LN))]);
xlabel('Longitude in  deg.'); ylabel('Latitude in deg.'); title(['X (-Bt) component at ' num2str(alt_min) ' km.']); 

figure;scatter(Longi(LN),90-CLati(LN),6,detrend(Bcrust(:,2)),'filled'); colormap(map); caxis([-2*std(dFesto) 2*std(dFesto)]); colorbar
hold on; plot(long,lat,'k', 'linewidth',2); xlim([min(Longi(LN)) max(Longi(LN))]); ylim([min(90-CLati(LN)) max(90-CLati(LN))]);
xlabel('Longitude in  deg.'); ylabel('Latitude in deg.'); title(['Y (Bp) East component at ' num2str(alt_min) ' km.']); 

figure;scatter(Longi(LN),90-CLati(LN),6,detrend(Bcrust(:,3)),'filled'); colormap(map); caxis([-2*std(dFesto) 2*std(dFesto)]); colorbar
hold on; plot(long,lat,'k', 'linewidth',2); xlim([min(Longi(LN)) max(Longi(LN))]); ylim([min(90-CLati(LN)) max(90-CLati(LN))]);
xlabel('Longitude in  deg.'); ylabel('Latitude in deg.'); title(['Z (-Br) Downward  component at ' num2str(alt_min) ' km.']); 

end

disp('Ended successfully...')

return

function B = Sph2rectangular(L0,C0,L,C,Bin,type)
%%% Vector transformation from spherical Earth to Local cartesian system
%%% [L0,C0] :  center coodinates (long, colatitude) in geocentric
%%% [L ,C ] : data geocentric coordinates (long, colatitude)
%%% Bin : input vector field in [X,Y,Z], where X is North, Y is East and Z
%%% is down in geocentric coordinates.
%%% B = [Xc,Yc,Zc] in the Cartesian reference frame with Xc North, Yc East
%%% and Zc down.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% E. THEBAULT, Centre National de la Recherche Scientifique (CNRS)
%%% University of Clermont Auvergne, IRD, OPGC, Laboratoire Magmas et
%%% Volcans, Clermont-Ferrand, France
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
L0 = L0*pi/180;
C0 = C0*pi/180;
L = L*pi/180;
C = C*pi/180;

if strcmp(type,'forward')
    %%% First rotation
    B1(:,1) = -Bin(:,1).*cos(C).*cos(L)-Bin(:,2).*sin(L)-Bin(:,3).*sin(C).*cos(L);
    B1(:,2) = -Bin(:,1).*cos(C).*sin(L)+Bin(:,2).*cos(L)-Bin(:,3).*sin(C).*sin(L);
    B1(:,3) =  Bin(:,1).*sin(C)                         -Bin(:,3).*cos(C);
    
    %%% Second rotation
    B(:,1) = -B1(:,1).*cos(C0).*cos(L0)-B1(:,2).*cos(C0).*sin(L0)+B1(:,3).*sin(C0);
    B(:,2) = -B1(:,1).*sin(L0)         +B1(:,2).*cos(L0);
    B(:,3) = -B1(:,1).*sin(C0).*cos(L0)-B1(:,2).*sin(C0).*sin(L0)-B1(:,3).*cos(C0);
    
elseif strcmp(type,'reverse')
%%% note: could be better written analytically  
B = zeros(size(Bin));

    M1 = [-cos(C0)*cos(L0) -cos(C0)*sin(L0)  sin(C0);
          -sin(L0)          cos(L0)                0;
          -sin(C0)*cos(L0) -sin(C0)*sin(L0) -cos(C0)];
    
    for i =1:size(L,1)
        %%% First Matrix
        M2 = [-cos(C(i)).*cos(L(i)) -sin(L(i))            -sin(C(i)).*cos(L(i));
              -cos(C(i)).*sin(L(i))  cos(L(i))            -sin(C(i)).*sin(L(i));
               sin(C(i))              0                    -cos(C(i))          ];
        
        B(i,:) = ((M1*M2)\Bin(i,:)')';
    end
end


return

function   X=PreCondLight_2_threshold(MATA,BKT,lmin)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% E. THEBAULT, Centre National de la Recherche Scientifique (CNRS)
%%% University of Clermont Auvergne, IRD, OPGC, Laboratoire Magmas et Volcans, Clermont-Ferrand, France

% 1) get the Matrix norm
DD=sqrt(diag(MATA));
V=diag(1./DD);

% 2) Preconditionning
MKT=MATA*V;
MK=V*MKT;
clear MATA MKT;
BK=V*BKT;
clear BKT MAT;

% 3) Singular value decomposition
[A,S,~]=svd(MK);
%%% since MK is positive real and symmetric B = A, no need to return B
clear MK

eigenvs = diag(S);
clear S

LambMax=max(eigenvs);
LambMin=min(eigenvs);
LambI=1./eigenvs;

% 4) Cancel out eigenvalues below threshold
Lsingul = eigenvs/LambMax < lmin;
LambI(Lsingul) = 0;
SI=diag(LambI);

% 5) Generalized inverse
MKi=A*SI*A';
disp(['Matrix initial   conditionning: ' num2str(LambMax/LambMin) ]);
disp(['Number of truncated eigenvalues: ' num2str(sum(Lsingul))]);
disp(['Matrix truncated conditionning: ' num2str(LambMax/min(eigenvs(logical(1-Lsingul))))]);
NPAR=MKi*BK;
X=V*NPAR;
return

function [New_Long,New_Colat] = cal2geoc(long_pole,colat_pole,...
					 long_cal,colat_cal);
% [LONG COLAT] = CAL2GEOC(LONG_POLE,COLAT_POLE,LONG_CAL,COLAT_CAL)
%  Rotation of the a spherical cap coordinate system to a geocentric reference frame
%  centered on long_pole, colat_pole, respectively the longitude and the 
%  colatitude of the new reference frame in degrees
%  version 1.0 02/02/23
%  E. THEBAULT, Centre National de la Recherche Scientifique (CNRS)
%  University of Clermont Auvergne, IRD, OPGC, Laboratoire Magmas et Volcans, Clermont-Ferrand, France

if nargin<4
   fprintf('Not enough input arguments');
   help cal2geoc
end

long_pole = long_pole*pi/180;
colat_pole = colat_pole*pi/180;

% First rotation
rotation1 = [ cos(long_pole) -sin(long_pole) 0;
              sin(long_pole)  cos(long_pole) 0;
              0               0              1 ];
% Second rotation

rotation2 = [ cos(colat_pole)   0     sin(colat_pole);
	      0                 1           0;
	     -sin(colat_pole)   0      cos(colat_pole)];

% Full rotation matrix
P = rotation1*rotation2;

if size(long_cal,2)<size(long_cal,1)
    long_cal = long_cal';
    colat_cal = colat_cal';
end

[xc, yc, zc] =  sph2cart(long_cal*pi/180,(90-colat_cal)*pi/180,...
		       ones(1,size(long_cal,2)));
W = [xc;yc;zc];

% new reference frame
W_geoc = P*W;

xn = W_geoc(1,:);
yn = W_geoc(2,:);
zn = W_geoc(3,:);

% back to spherical
[New_Long New_Colat Rho] = cart2sph(xn,yn,zn);

New_Long = New_Long'*180/pi;
New_Colat = (pi/2-New_Colat')*180/pi;

return


function [New_Long,New_Colat] = geoc2cal(long_pole,colat_pole,...
    long_geo,colat_geo)
% [LONG_CAl COLAT_CAL] = GEOC2CAL(LONG_POLE,COLAT_POLE,LONG,COLAT)
%  Rotation of the geocentric coordinate system to a cap reference frame
%  centered on long_pole, colat_pole, respectively the longitude and the 
%  colatitude of the new reference frame in degrees
%  version 1.0 02/02/23
%  E. THEBAULT, Centre National de la Recherche Scientifique (CNRS)
%  University of Clermont Auvergne, IRD, OPGC, Laboratoire Magmas et Volcans, Clermont-Ferrand, France
% =================================================================

if nargin<4
    fprintf('Not enough input arguments');
    help cal2geoc
end

long_pole = long_pole*pi/180;
colat_pole = colat_pole*pi/180;

% First rotation
rotation1 = [ cos(long_pole)   -sin(long_pole)  0;
    sin(long_pole)   cos(long_pole)  0;
    0                 0              1 ];

% Second rotation

rotation2 = [ cos(colat_pole)   0    sin(colat_pole);
    0            1           0;
    -sin(colat_pole)   0    cos(colat_pole)];

% Full rotation matrix
P = rotation1*rotation2;

if size(long_geo,2)<size(long_geo,1)
    long_geo = long_geo';
    colat_geo = colat_geo';
end

[xg,yg,zg] =  sph2cart(long_geo*pi/180,(90-colat_geo)*pi/180,...
    ones(1,size(long_geo,2)));
W = [xg;yg;zg];

% new reference frame
W_cal = P'*W;

xn = W_cal(1,:);
yn = W_cal(2,:);
zn = W_cal(3,:);

% back to spherical
[New_Long,New_lat,~] = cart2sph(xn,yn,zn);

New_Long = New_Long'*180/pi;
New_Colat = (pi/2-New_lat')*180/pi;

return;


function A =prj_lamA(points,type)
%%% Simple Projection assuming a spherical Earth
%%% E. THEBAULT
%%% Centre National de la Recherche Scientifique
%%% Universite Clermont Auvergne, IRD, OPGC, Laboratoire Magmas et Volcans, Clermont-Ferrand, France
%%% Revised and optimized in March 2023

Rt = 6371.2;
if strcmp(type,'forward')
    long = points(:,1);
    colat= points(:,2);
    %alt  = points(:,3);
    % =================== Azimutal Projection ====================
    r = 2.*Rt.*sin(colat.*pi/360);
    xp= r.*cos(long.*pi/180);
    yp =r.*sin(long.*pi/180);
    zp = Rt.*(1-cos(colat*pi/180));
    
    A=[xp yp zp];
    
    
else
    %%% Reverse Lambert Projection assuming a spherical Earth
    xp = points(:,1);
    yp = points(:,2);
    %alt = points(:,3); %%% alt is altitude with respect to the Earth's surface
    
    % =================== Spherical Projection ====================
    zp = (Rt-sqrt(Rt^2-(yp.^2+xp.^2)));
    long = atan2(yp,xp)*180/pi;
    r = sqrt(xp.^2+yp.^2+zp.^2);
    colat = 2*asin(zp./r)*180/pi;
    A = [long colat];
end


return

function [MAT,INDX, MX,MY,MZ] = design_RHA(x,y,z,CoreField,DXp,DYp,NMAX)
%%% [MX,MY,MZ,MAT] = Design_RHA(x,y,z,DXp,DYp,DZp,NMAX)
%%% Design matrix for the Rectangular Harmonic Analysis
%%% E. THEBAULT
%%% Centre National de la Recherche Scientifique
%%% Universite Clermont Auvergne, IRD, OPGC, Laboratoire Magmas et Volcans, Clermont-Ferrand, France
%%% Revised and optimized in March 2023
%%% Revised in september 2023 to append n within INDX

%%% Initialization

Xcore = CoreField(:,1);
Ycore = CoreField(:,2);
Zcore = CoreField(:,3);
Fcore = CoreField(:,4);

Cpx = 2*pi/DXp;
Cpy = 2*pi/DYp;

xp = Cpx*x;
yp = Cpy*y;
zp = z;

npx = size(x,1);



if nargout > 2
    size_n = 4*(NMAX+1)^2;
    MAT = zeros(npx,size_n);
    INDX = zeros(size_n,3);
    
    MZ = MAT;
    MX = MAT;
    MY = MAT;
    iter = 1;
    nn = [];
    
    for ii=0:NMAX
        for jj=0:NMAX
            
            n = sqrt((ii*Cpx)^2+(jj*Cpy)^2);
            
            %%% radial function
            R = exp(n*(zp));
            
            % trigonometric functions for the potential
            A1 =  sin(ii*xp).*sin(jj*yp).*R;
            A2 =  cos(ii*xp).*cos(jj*yp).*R;
            A3 =  cos(ii*xp).*sin(jj*yp).*R;
            A4 =  sin(ii*xp).*cos(jj*yp).*R;
            
            % Basis function for the anomaly field projected along the ambiant
            % field.
            % Basis function for the North component
            MX(:,iter)   =   Cpx.*ii*A3;
            MX(:,iter+1) =  -Cpx.*ii*A4;
            MX(:,iter+2) =  -Cpx.*ii*A1;
            MX(:,iter+3) =   Cpx.*ii*A2;
            
            % Basis function for the East component
            MY(:,iter)   =   Cpy.*jj*A4;
            MY(:,iter+1) =  -Cpy.*jj*A3;
            MY(:,iter+2) =   Cpy.*jj*A2;
            MY(:,iter+3) =  -Cpy.*jj*A1;
            
            % Basis function for the donward vertical component
            MZ(:,iter)   =     n.*A1;
            MZ(:,iter+1) =     n.*A2;
            MZ(:,iter+2) =     n.*A3;
            MZ(:,iter+3) =     n.*A4;
            
            MAT(:,iter)   =  Cpx.*ii*A3.*Xcore./Fcore + Cpy.*jj*A4.*Ycore./Fcore + n.*A1.*Zcore./Fcore;
            MAT(:,iter+1) = -Cpx.*ii*A4.*Xcore./Fcore - Cpy.*jj*A3.*Ycore./Fcore + n.*A2.*Zcore./Fcore;
            MAT(:,iter+2) = -Cpx.*ii*A1.*Xcore./Fcore + Cpy.*jj*A2.*Ycore./Fcore + n.*A3.*Zcore./Fcore;
            MAT(:,iter+3) =  Cpx.*ii*A2.*Xcore./Fcore - Cpy.*jj*A1.*Ycore./Fcore + n.*A4.*Zcore./Fcore;
            
            
            INDX(iter:iter+3,1:3) = repmat([ii jj n],4,1);
            iter = iter + 4;
        end
    end
else
    iter = 1;
    %%% only the intensity matrix is needed
    size_n = 4*(NMAX+1)^2;
    MAT = zeros(npx,size_n);
    INDX = zeros(size_n,3);
    
    for ii=0:NMAX
        for jj=0:NMAX
            
            n = sqrt((ii*Cpx)^2+(jj*Cpy)^2);
            R = exp(n*(zp));
            
            % trigonometric functions for the potential
            A1 =  sin(ii*xp).*sin(jj*yp).*R;
            A2 =  cos(ii*xp).*cos(jj*yp).*R;
            A3 =  cos(ii*xp).*sin(jj*yp).*R;
            A4 =  sin(ii*xp).*cos(jj*yp).*R;
            
            % Basis function for the anomaly field projected along the ambiant
            % field.
            MAT(:,iter)   =  Cpx.*ii*A3.*Xcore./Fcore + Cpy.*jj*A4.*Ycore./Fcore + n.*A1.*Zcore./Fcore;
            MAT(:,iter+1) = -Cpx.*ii*A4.*Xcore./Fcore - Cpy.*jj*A3.*Ycore./Fcore + n.*A2.*Zcore./Fcore;
            MAT(:,iter+2) = -Cpx.*ii*A1.*Xcore./Fcore + Cpy.*jj*A2.*Ycore./Fcore + n.*A3.*Zcore./Fcore;
            MAT(:,iter+3) =  Cpx.*ii*A2.*Xcore./Fcore - Cpy.*jj*A1.*Ycore./Fcore + n.*A4.*Zcore./Fcore;
            
            INDX(iter:iter+3,1:3) = repmat([ii jj n],4,1);
            iter = iter + 4;
        end
    end
    
        
end
return


function B_MF  = ESTIMATE_CORE_FIELD(t, r, theta, phi, GH_Splines)
%%% function B_MF  = ESTIMATE_CORE_FIELD(r, theta, phi, GH_Splines)
%%% FUNCTIONS COMPUTING THE EARTH's MAIN FIELD
%%% Input  = [RADIUS(km) t(MJD) COLAT(rad) LONG(rad) Splines]; MJD is Modified
%%% Output = [B_r_core B_theta_core B_Phi_core F_core]
%%% GH_Splines containing the SH Gauss coefficients as splines
%%% functions provided by the CHAOS format: http://www.spacecenter.dk/files/magnetic-models/CHAOS-7/
%%% E. THEBAULT
%%% Centre National de la Recherche Scientifique
%%% Université Clermont Auvergne, IRD, OPGC, Laboratoire Magmas et Volcans, Clermont-Ferrand, France
%%% V2.0, March 2023
a = 6371.2;
r = r/a;
maxdeg_MF = sqrt(GH_Splines.dim+1)-1;

%%% compute basis functions in spherical harmonics
cos_theta = cos(theta);
sin_theta = sin(theta);
N_data    = length(theta);

A_r       = zeros(N_data, GH_Splines.dim);
A_theta   = zeros(N_data, GH_Splines.dim);
A_phi     = zeros(N_data, GH_Splines.dim);

k=0;

for n = 1:maxdeg_MF
    
    r_n       = r.^(-(n+2));
    Pnm = legendre(n, cos_theta, 'sch')';
    dPnm(:,n+1) =  (sqrt(n/2).*Pnm(:,n));
    dPnm(:,1) = -sqrt(n*(n+1)/2.).*Pnm(:,2);
    
    if n > 1
        dPnm(:,2)=(sqrt(2*(n+1)*n).*Pnm(:,1)-sqrt((n+2)*(n-1)).*Pnm(:,3))/2;
    end
    
    for m = 2:n-1
        dPnm(:,m+1)=(sqrt((n+m)*(n-m+1)).*Pnm(:,m)-sqrt((n+m+1)*(n-m)).*Pnm(:,m+2))/2;
    end
    
    if n == 1
        dPnm(:,2) = sqrt(2)*dPnm(:,2);
    end
    
    
    for m = 0:n
        cos_phi   = cos(m*phi);
        sin_phi   = sin(m*phi);
        
        if m == 0
            k = k+1;
            A_r(:,k)       =  (n+1).*r_n(:).*Pnm(:,1);
            A_theta(:,k)   = -r_n(:).*dPnm(:,1);
            A_phi(:,k)     =  r_n(:)*0;
        else
            k = k+1;
            A_r(:,k)       =  (n+1).*r_n(:).*cos_phi.*Pnm(:,m+1);
            A_theta(:,k)   = -r_n(:).*cos_phi.*dPnm(:,m+1);
            A_phi(:,k)     =  r_n(:).*m.*sin_phi.*Pnm(:,m+1)./sin_theta;
            
            k = k+1;
            A_r(:,k)       =  (n+1).*r_n(:).*sin_phi.*Pnm(:,m+1);
            A_theta(:,k)   = -r_n(:).*sin_phi.*dPnm(:,m+1);
            A_phi(:,k)     = -r_n(:).*m.*cos_phi.*Pnm(:,m+1)./sin_theta;
        end
    end
    
end

%%% Now estimate the coefficients at epoch t and multiply
g_i = ppval(t', GH_Splines);
B_MF = [sum(A_r.*g_i',2) sum(A_theta.*g_i',2) sum(A_phi.*g_i',2)];
B_MF(:,4) = sqrt(B_MF(:,1).^2+B_MF(:,2).^2+B_MF(:,3).^2);
return


function map = MyColorMap()
map =[
    0.2510         0    0.2510
    0.2301         0    0.3134
    0.2092         0    0.3758
    0.1882         0    0.4383
    0.1673         0    0.5007
    0.1464         0    0.5631
    0.1255         0    0.6255
    0.1046         0    0.6879
    0.0837         0    0.7503
    0.0627         0    0.8127
    0.0418         0    0.8752
    0.0209         0    0.9376
         0         0    1.0000
    0.0111    0.1111    1.0000
    0.0222    0.2222    1.0000
    0.0333    0.3333    1.0000
    0.0444    0.4444    1.0000
    0.0556    0.5556    1.0000
    0.0667    0.6667    1.0000
    0.0778    0.7778    1.0000
    0.0889    0.8889    1.0000
    0.1000    1.0000    1.0000
    0.1900    1.0000    1.0000
    0.2800    1.0000    1.0000
    0.3700    1.0000    1.0000
    0.4600    1.0000    1.0000
    0.5500    1.0000    1.0000
    0.6400    1.0000    1.0000
    0.7300    1.0000    1.0000
    0.8200    1.0000    1.0000
    0.9100    1.0000    1.0000
    1.0000    1.0000    1.0000
    1.0000    1.0000    0.9000
    1.0000    1.0000    0.8000
    1.0000    1.0000    0.7000
    1.0000    1.0000    0.6000
    1.0000    1.0000    0.5000
    1.0000    1.0000    0.4000
    1.0000    1.0000    0.3000
    1.0000    1.0000    0.2000
    1.0000    1.0000    0.1000
    1.0000    1.0000         0
    1.0000    0.9091         0
    1.0000    0.8182         0
    1.0000    0.7273         0
    1.0000    0.6364         0
    1.0000    0.5455         0
    1.0000    0.4545         0
    1.0000    0.3636         0
    1.0000    0.2727         0
    1.0000    0.1818         0
    1.0000    0.0909         0
    1.0000         0         0
    1.0000         0    0.0909
    1.0000         0    0.1818
    1.0000         0    0.2727
    1.0000         0    0.3636
    1.0000         0    0.4545
    1.0000         0    0.5455
    1.0000         0    0.6364
    1.0000         0    0.7273
    1.0000         0    0.8182
    1.0000         0    0.9091
    1.0000         0    1.0000];
return

